# -*- coding: utf-8 -*-
from __future__ import unicode_literals


"""
pt-br custom locale file.
"""

translations = {
    # Relative time
    "after": "após {0}",
    "before": "{0} atrás",
    # Date formats
    "date_formats": {
        "LTS": "HH:mm:ss",
        "LT": "HH:mm",
        "LLLL": "dddd, D [de] MMMM [de] YYYY [às] HH:mm",
        "LLL": "D [de] MMMM [de] YYYY [às] HH:mm",
        "LL": "D [de] MMMM [de] YYYY",
        "L": "DD/MM/YYYY",
    },
}
